'use strict';

var _objectSpread = require('@babel/runtime/helpers/objectSpread2');
var httpHelpers = require('@toruslabs/http-helpers');
var bignumber_js = require('bignumber.js');
var constants = require('../utils/constants.js');
var conversionUtils = require('../utils/conversionUtils.js');

function normalizeGWEIDecimalNumbers(n) {
  const numberAsWEIHex = conversionUtils.decGWEIToHexWEI(new bignumber_js.BigNumber(n));
  const numberAsGWEI = conversionUtils.hexWEIToDecGWEI(numberAsWEIHex);
  return numberAsGWEI;
}
async function fetchGasEstimates(url) {
  const estimates = await httpHelpers.get(url);
  const normalizedEstimates = {
    estimatedBaseFee: normalizeGWEIDecimalNumbers(estimates.estimatedBaseFee).toString(10),
    low: _objectSpread(_objectSpread({}, estimates.low), {}, {
      suggestedMaxPriorityFeePerGas: normalizeGWEIDecimalNumbers(estimates.low.suggestedMaxPriorityFeePerGas).toString(10),
      suggestedMaxFeePerGas: normalizeGWEIDecimalNumbers(estimates.low.suggestedMaxFeePerGas).toString(10)
    }),
    medium: _objectSpread(_objectSpread({}, estimates.medium), {}, {
      suggestedMaxPriorityFeePerGas: normalizeGWEIDecimalNumbers(estimates.medium.suggestedMaxPriorityFeePerGas).toString(10),
      suggestedMaxFeePerGas: normalizeGWEIDecimalNumbers(estimates.medium.suggestedMaxFeePerGas).toString(10)
    }),
    high: _objectSpread(_objectSpread({}, estimates.high), {}, {
      suggestedMaxPriorityFeePerGas: normalizeGWEIDecimalNumbers(estimates.high.suggestedMaxPriorityFeePerGas).toString(10),
      suggestedMaxFeePerGas: normalizeGWEIDecimalNumbers(estimates.high.suggestedMaxFeePerGas).toString(10)
    })
  };
  return normalizedEstimates;
}
async function fetchGasEstimatesViaEthFeeHistory(provider) {
  const noOfBlocks = 10;
  const newestBlock = "latest";
  // get the 10, 50 and 95th percentile of the tip fees from the last 10 blocks
  const percentileValues = [10, 50, 95];
  const feeHistory = await provider.request({
    method: "eth_feeHistory",
    params: [noOfBlocks, newestBlock, percentileValues]
  });
  // this is in hex wei
  const finalBaseFeePerGas = feeHistory.baseFeePerGas[feeHistory.baseFeePerGas.length - 1];
  // this is in hex wei
  const rewardList = feeHistory.reward.filter(reward => reward[0] !== "0x0" || reward[1] !== "0x0" || reward[2] !== "0x0");
  const numberOfRewards = rewardList.length || 1;
  const priorityFeeCalcs = rewardList.reduce((acc, curr) => {
    return {
      slow: acc.slow.plus(new bignumber_js.BigNumber(curr[0], 16)),
      average: acc.average.plus(new bignumber_js.BigNumber(curr[1], 16)),
      fast: acc.fast.plus(new bignumber_js.BigNumber(curr[2], 16))
    };
  }, {
    slow: new bignumber_js.BigNumber(0),
    average: new bignumber_js.BigNumber(0),
    fast: new bignumber_js.BigNumber(0)
  });
  return {
    estimatedBaseFee: conversionUtils.hexWEIToDecGWEI(finalBaseFeePerGas).toString(10),
    high: {
      maxWaitTimeEstimate: 30000,
      minWaitTimeEstimate: 15000,
      suggestedMaxFeePerGas: conversionUtils.hexWEIToDecGWEI(priorityFeeCalcs.fast.div(numberOfRewards).plus(finalBaseFeePerGas).toString(16)).toString(),
      suggestedMaxPriorityFeePerGas: conversionUtils.hexWEIToDecGWEI(priorityFeeCalcs.fast.div(numberOfRewards).toString(16)).toString()
    },
    medium: {
      maxWaitTimeEstimate: 45000,
      minWaitTimeEstimate: 15000,
      suggestedMaxFeePerGas: conversionUtils.hexWEIToDecGWEI(priorityFeeCalcs.average.div(numberOfRewards).plus(finalBaseFeePerGas).toString(16)).toString(),
      suggestedMaxPriorityFeePerGas: conversionUtils.hexWEIToDecGWEI(priorityFeeCalcs.average.div(numberOfRewards).toString(16)).toString()
    },
    low: {
      maxWaitTimeEstimate: 60000,
      minWaitTimeEstimate: 15000,
      suggestedMaxFeePerGas: conversionUtils.hexWEIToDecGWEI(priorityFeeCalcs.slow.div(numberOfRewards).plus(finalBaseFeePerGas).toString(16)).toString(),
      suggestedMaxPriorityFeePerGas: conversionUtils.hexWEIToDecGWEI(priorityFeeCalcs.slow.div(numberOfRewards).toString(16)).toString()
    }
  };
}
/**
 * Hit the legacy MetaSwaps gasPrices estimate api and return the low, medium
 * high values from that API.
 */
async function fetchLegacyGasPriceEstimates(url) {
  const result = await httpHelpers.get(url, {
    referrer: url,
    referrerPolicy: "no-referrer-when-downgrade",
    method: "GET"
  });
  // this returns decimal gwei
  return {
    low: result.SafeGasPrice,
    medium: result.ProposeGasPrice,
    high: result.FastGasPrice
  };
}
async function fetchEthGasPriceEstimate(provider) {
  const gasPrice = await provider.request({
    method: constants.METHOD_TYPES.ETH_GET_GAS_PRICE
  });
  return {
    gasPrice: conversionUtils.hexWEIToDecGWEI(gasPrice).toString()
  };
}
function calculateTimeEstimate(maxPriorityFeePerGas, maxFeePerGas, gasFeeEstimates) {
  // all are in dec gwei
  const {
    low,
    medium,
    high,
    estimatedBaseFee
  } = gasFeeEstimates;
  const maxPriorityFeePerGasInWEI = new bignumber_js.BigNumber(conversionUtils.decGWEIToHexWEI(new bignumber_js.BigNumber(maxPriorityFeePerGas)), 16);
  const maxFeePerGasInWEI = new bignumber_js.BigNumber(conversionUtils.decGWEIToHexWEI(new bignumber_js.BigNumber(maxFeePerGas)), 16);
  const estimatedBaseFeeInWEI = new bignumber_js.BigNumber(conversionUtils.decGWEIToHexWEI(new bignumber_js.BigNumber(estimatedBaseFee)), 16);
  const effectiveMaxPriorityFee = bignumber_js.BigNumber.min(maxPriorityFeePerGasInWEI, maxFeePerGasInWEI.minus(estimatedBaseFeeInWEI));
  const lowMaxPriorityFeeInWEI = new bignumber_js.BigNumber(conversionUtils.decGWEIToHexWEI(new bignumber_js.BigNumber(low.suggestedMaxPriorityFeePerGas)), 16);
  const mediumMaxPriorityFeeInWEI = new bignumber_js.BigNumber(conversionUtils.decGWEIToHexWEI(new bignumber_js.BigNumber(medium.suggestedMaxPriorityFeePerGas)), 16);
  const highMaxPriorityFeeInWEI = new bignumber_js.BigNumber(conversionUtils.decGWEIToHexWEI(new bignumber_js.BigNumber(high.suggestedMaxPriorityFeePerGas)), 16);
  let lowerTimeBound;
  let upperTimeBound;
  if (effectiveMaxPriorityFee.lt(lowMaxPriorityFeeInWEI)) {
    lowerTimeBound = null;
    upperTimeBound = "unknown";
  } else if (effectiveMaxPriorityFee.gte(lowMaxPriorityFeeInWEI) && effectiveMaxPriorityFee.lt(mediumMaxPriorityFeeInWEI)) {
    lowerTimeBound = low.minWaitTimeEstimate;
    upperTimeBound = low.maxWaitTimeEstimate;
  } else if (effectiveMaxPriorityFee.gte(mediumMaxPriorityFeeInWEI) && effectiveMaxPriorityFee.lt(highMaxPriorityFeeInWEI)) {
    lowerTimeBound = medium.minWaitTimeEstimate;
    upperTimeBound = medium.maxWaitTimeEstimate;
  } else if (effectiveMaxPriorityFee.eq(highMaxPriorityFeeInWEI)) {
    lowerTimeBound = high.minWaitTimeEstimate;
    upperTimeBound = high.maxWaitTimeEstimate;
  } else {
    lowerTimeBound = 0;
    upperTimeBound = high.maxWaitTimeEstimate;
  }
  return {
    lowerTimeBound,
    upperTimeBound
  };
}

exports.calculateTimeEstimate = calculateTimeEstimate;
exports.fetchEthGasPriceEstimate = fetchEthGasPriceEstimate;
exports.fetchGasEstimates = fetchGasEstimates;
exports.fetchGasEstimatesViaEthFeeHistory = fetchGasEstimatesViaEthFeeHistory;
exports.fetchLegacyGasPriceEstimates = fetchLegacyGasPriceEstimates;
exports.normalizeGWEIDecimalNumbers = normalizeGWEIDecimalNumbers;
